import sys
import re
import enum

# Easy error printing
def eprint(error: str) -> None:
    print(error, file=sys.stderr)

class Style(enum.Enum):
    NONE = enum.auto()
    HEADING = enum.auto()
    TAG = enum.auto()
    CARD = enum.auto()
    HIGHLIGHT = enum.auto()

class TextElement:
    def __init__(self, txt: str, style: Style, raw: bool = False) -> None:
        self.txt = txt
        self.style = style
        if not raw:
            self.sanitize()
    def sanitize(self) -> None:
        self.txt = re.sub(r'\\([^{}]|$)', r'\\textbackslash{}\1', self.txt, 0,
                re.M)
        self.txt = self.txt.replace('&', '\\&')
        self.txt = self.txt.replace('%', '\\%')
        self.txt = self.txt.replace('$', '\\$')
        self.txt = self.txt.replace('#', '\\#')
        self.txt = self.txt.replace('_', '\\_')
        self.txt = self.txt.replace('~', '\\texttildelow{}')
        self.txt = self.txt.replace('^', '\\textasciicircum{}')
    def __repr__(self) -> str:
        return repr(self.__dict__)

class Element:
    def __init__(self, itext: str) -> None:
        self.text = []
        if re.search(r'^#[^#]', itext):
            self.text.append(TextElement(itext[1:].lstrip(), Style.HEADING))
            self.type = Style.HEADING
            return
        elif re.search(r'^##[^#]', itext):
            self.text.append(TextElement(itext[2:].lstrip(), Style.TAG))
            self.type = Style.TAG
            return
        elif re.search(r'^###', itext):
            itext = itext[2:]
        self.type = Style.CARD
        f = re.search(r'(\s|^)\{', itext)
        if not f:
            self.text.append(TextElement(itext, Style.CARD))
            return
        i = 0
        while f:
            self.text.append(TextElement(itext[i:i + f.end() - 1], Style.CARD))
            i += f.end() - 1
            t = re.search(r'(^|[^\\])\{', itext[i:])
            if t:
                t2 = re.search(r'(^|[^\\])\{\{', itext[i:])
                if t2 and t2.start == t.start:
                    m = re.search(r'(^|[^\\])\{\{(.*?)(^|[^\\])\}\}', itext[i:],
                            re.S | re.M)
                    if not m:
                        eprint('Unclosed {{')
                        sys.exit(1)
                    self.text.append(TextElement(m[2] + m[3], Style.CARD, True))
                    i += m.end()
                else:
                    m = re.search(r'(^|[^\\])\{(.*?)(^|[^\\])\}', itext[i:],
                            re.S | re.M)
                    if not m:
                        eprint('Unclosed {')
                        sys.exit(1)
                    self.text.append(TextElement(m[2] + m[3], Style.HIGHLIGHT))
                    i += m.end()
            f = re.search(r'(^|[^\\])\{', itext[i:])
        self.text.append(TextElement(itext[i:], Style.CARD))
    def __repr__(self) -> str:
        return repr(self.__dict__)

class LatexData:
    PREAMBLE = r'''\documentclass[10pt]{article}
\usepackage[letterpaper]{geometry}
\geometry{top=0.8in,left=0.8in,bottom=0.8in,right=0.8in}
\usepackage{fontspec}
\setmainfont[Ligatures=TeX]{Noto Sans}
\pagestyle{empty}
\usepackage{anyfontsize}
\usepackage[utf8]{inputenc}
\usepackage{color,soul}
\begin{document}
\thispagestyle{empty}
\setlength{\parindent}{0in}
'''
    STYLES = {Style.NONE:
                  r'',
              Style.HEADING:
                  r'\fontsize{20}{20}\selectfont{\centering\textbf{',
              Style.TAG:
                  r'\fontsize{10}{10}\selectfont\textbf{',
              Style.CARD:
                  r'\fontsize{6}{6}\selectfont{}',
              Style.HIGHLIGHT:
                  r'\fontsize{10}{10}\selectfont\textbf{\hl{'}
    STYLEENDS = {Style.NONE:
                     r'',
                 Style.HEADING:
                     r'}\par}',
                 Style.TAG:
                     r'}',
                 Style.CARD:
                     r'',
                 Style.HIGHLIGHT:
                     r'}}'}
    POSTAMBLE = r'\end{document}'

def main() -> int:
    if len(sys.argv) < 3:
        eprint('Usage: ' + sys.argv[0] + ' <input file> <output file>')
        return 1
    ifile = []
    with open(sys.argv[1], 'r') as f:
        ifile = f.readlines()
    elements = []
    lno = 0
    while lno < len(ifile):
        par = ''
        while lno < len(ifile) and ifile[lno] != '\n':
            par += ifile[lno]
            lno += 1
        elements.append(Element(par))
        lno += 1
    ofile = LatexData.PREAMBLE
    cstyle: Style = Style.NONE
    print(elements)
    for i, element in enumerate(elements):
        if i == 0:
            pass
        elif element.type == Style.HEADING:
            ofile += '\\newpage\n'
        elif elements[i - 1].type == Style.HEADING:
            pass
        elif element.type == Style.TAG:
            ofile += LatexData.STYLEENDS[cstyle] + \
                    '\\par\\unskip\\vspace{5pt}\n' + LatexData.STYLES[cstyle]

        for te in element.text:
            if te.style != cstyle:
                if (cstyle != Style.NONE):
                    ofile += LatexData.STYLEENDS[cstyle]
                ofile += LatexData.STYLES[te.style]
                cstyle = te.style
            ofile += te.txt
        if (i < len(elements) - 1 and elements[i + 1].type != Style.HEADING
                and element.type != Style.HEADING):
            ofile += '\\newline\n'
    ofile += LatexData.STYLEENDS[cstyle] + LatexData.POSTAMBLE
    with open(sys.argv[2], 'w') as f:
        f.write(ofile)
    return 0
